/*
  definition of graphs and their iterations
*/

#include <map>
#include <vector>
#include <iostream>
#include <algorithm>
#include <utility>
#include <cassert>
#include <ext/hash_set>
#include <ext/hash_map>
#include <cstdlib>
using namespace std;
using namespace __gnu_cxx;

#include "updatable_heap.cpp"

__DEBUG(int max_N_size = 0;)
typedef vector< vector<edge> > adj;

struct graph {
  adj e;

  graph(istream &is);
  ~graph();

  void print() const;
  void search() const;
  void pushPath(const path &p, WL *wl, const hash_map<val, val, val_hash_t, val_eq_t> &n) const;
};

graph::graph(istream &is)
{
  int size; is >> size;
  e.resize(size);
  while (!is.eof()) {
    edge e1; is >> e1;
    this->e[e1.in].push_back(e1);
  }
}

graph::~graph() {}

// updating the heap
void inline graph::pushPath(const path &p, WL *wl, const hash_map<val, val, val_hash_t, val_eq_t> &n) const
{
  if (unnecessary(p.v)) return;
  if (n.find(p.v) != n.end()) {return;}

  wl->push_or_update(p);
}

void graph::print() const
{
  for (adj::const_iterator it = e.begin(); it != e.end(); it++) {
	cout << (it-e.begin()) << " => { ";
	for(vector<edge>::const_iterator it2 = it->begin(); 
		it2 != it->end(); it2++) {
	  cout << it2->out << "(" << it2->w << ") "; 
	}
	cout << "}" << endl;
  }
}

// Dijkstra search (with some modification)
void graph::search() const
{
  hash_map<val, val, val_hash_t, val_eq_t> n(1<<20);
  path ninf; ninf.v = val::ninf();
  WL wl(ninf); //make a heap with a "guard"

  for (adj::const_iterator it = e.begin(); it != e.end(); it++) {
	for(vector<edge>::const_iterator it2 = it->begin(); 
		it2 != it->end(); it2++) {
	  pushPath(path().extend(*it2), &wl, n); // add all edges to the heap
	}
  }
  while (!wl.empty()) {
    path p = wl.pop_minimum(); 
    if (constraint(p.v)) { 
	  // an optimal path is found
	  cout << "weight = " << p.weight() << endl;
	  vector<node> r;
	  r.push_back(p.v.n);
	  for (val v = p.prev; v.n != -1; v = n[v]) r.push_back(v.n);
	  cout << "[ ";
	  for (vector<node>::const_reverse_iterator it = r.rbegin(); 
		   it != r.rend(); it++ ) cout << *it << " ";
	  cout << "]" << endl;
	  return; 
	} 

    n[p.v] = p.prev;
    __DEBUG(if ((max_N_size = max(max_N_size, (int)n.size())) % 10000 == 0) cout << "max_N_size >= " << max_N_size++ << endl; );
    for (vector<edge>::const_iterator it = e[p.destination()].begin(); 
		 it != e[p.destination()].end(); it++) {
      pushPath(p.extend(*it), &wl, n);
    }
  }
  cout << "NO SOLUTION" << endl;
  return; // there is no feasible path.
}
