/*
  a heap with hashmap:
  we can directly access the node having the state, and update it.
*/


#include <map>
#include <vector>
#include <iostream>
#include <algorithm>
#include <utility>
#include <cassert>
#include <ext/hash_set>
#include <ext/hash_map>
#include <functional>
using namespace std;
using namespace __gnu_cxx;

__DEBUG(int max_heap = 0;)
__DEBUG(int indexing_size = 0;)

class WL {
public:
  WL(const path &ninf) : indexing(1<<20) { heap.push_back(new pair<path, int>(ninf, 0)); }
  ~WL() { __DEBUG(cerr << "max heap size = " << max_heap << endl;) }

  path pop_minimum();
  void push_or_update(path p);
  bool empty() const { return heap.size()==1; }

private:
  vector<pair<path, int>*> heap;
  hash_map<val, pair<path, int>*, val_hash_t, val_eq_t> indexing;
  
  void heap_upward(unsigned int index, pair<path, int>* p);
  void heap_downward(unsigned int index, pair<path, int>* p);
};

path WL::pop_minimum()
{
  path p = heap[1]->first; delete heap[1];
  pair<path, int>* last = heap.back(); heap.pop_back();
  heap_downward(1, last);
  indexing.erase(indexing.find(p.v));
  return p;
}

void WL::push_or_update(path p)
{
  hash_map<val, pair<path, int>*, val_hash_t, val_eq_t>::iterator it = indexing.find(p.v);
  if (it == indexing.end()) {
    pair<path, int>* added = new pair<path, int>(p, heap.size());
    indexing[p.v] = added;
    heap.push_back(NULL);
    heap_upward(heap.size() - 1, added);
    __DEBUG(if ((indexing_size = max(indexing_size, (int)indexing.size())) % 10000 == 0) cout << "indexing_size >= " << indexing_size++ << endl; )
    __DEBUG(max_heap = max(max_heap, (int)heap.size()); if (max_heap % 10000 == 0) cout << "max_heap_size >= " << max_heap++ << endl; )
  } else {
    if (it->second->first.weight() > p.weight()) {
      it->second->first = p;
      heap_upward(it->second->second, it->second);
    }
  }
}

void WL::heap_upward(unsigned int index, pair<path, int>* p)
{
  while (true) {
    unsigned int parent = index / 2;
    if (heap[parent]->first < p->first) break;
    heap[index] = heap[parent];
    heap[index]->second = index;
    index /= 2;
  }
  heap[index] = p;
  p->second = index;
}

void WL::heap_downward(unsigned int index, pair<path, int>* p)
{
  while (true) {
    unsigned int left = index * 2;
    unsigned int right = index * 2 + 1;
    if (left >= heap.size()) {
      break;
    } else if (right >= heap.size()) {
      if (p->first < heap[left]->first) break;
      heap[index] = heap[left]; heap[index]->second = index;
      index = left;
    } else if (heap[left]->first < heap[right]->first) {
      if (p->first < heap[left]->first) break;
      heap[index] = heap[left]; heap[index]->second = index;
      index = left;
    } else {
      if (p->first < heap[right]->first) break;
      heap[index] = heap[right]; heap[index]->second = index;
      index = right;
    }
  }
  heap[index] = p;
  p->second = index;
}
